from collections import defaultdict
from functools import reduce
from pathlib import Path
from typing import Any, Dict, Optional

from funcy import compose, update_in
import yaml


def compile_schema(spec_dir: Optional[str] = None) -> Dict[str, Any]:
    """Load schema from filesystem structure, determined by ``spec_dir`` attribute.

    :param spec_dir: root directory relative to the schema
    :return: swagger schema compatible with Connexion framework
    """
    defaults = _compile(Path(__file__).parent / "schemas")  # type: ignore

    if not spec_dir:
        return defaults

    applevel = _compile(spec_dir)

    def merge_values(a, b):
        return b if not (isinstance(a, dict) and isinstance(b, dict)) else {**a, **b}

    common = {k: merge_values(defaults[k], applevel[k]) for k in set(defaults) & set(applevel)}

    return _filter_null_keys({**defaults, **applevel, **common})


def _compile(spec_dir: str) -> Dict[str, Any]:
    root = Path(spec_dir)

    if not (root.exists() and root.is_dir()):
        raise ValueError(f"Directory {root} does not exist.")

    def load_spec(path):
        """Load partial schema spec from a *.yml file.

        :param path: A filesystem path
        :return: Return tuple of two items:
                 1) relative filesystem path parts serving as a key path for a nested dictionary
                 2) partial schema spec loaded from a *.yml file
        """
        with path.open() as f:
            spec = yaml.safe_load(f)

        return path.relative_to(root).parent.parts, spec  # type: ignore

    def update_value(acc, item):
        path, spec = item
        return update_in(acc, path, lambda val: {**val, **spec}, default={})

    return compose(
        lambda specs: reduce(update_value, specs, {}),
        # ↑ Build the dictionary
        lambda paths: map(load_spec, paths),
        # ↑ Load specs from fs paths
    )(root.glob("**/*.yml"))


def _filter_null_keys(d: Dict[str, Any]) -> Dict[str, Any]:
    """Recursively remove all keys from a dictionary that have None value.

    :param d: input dictionary
    :return: output dictionary with null keys removed
    """
    target: Dict[str, Any] = defaultdict(dict)

    for key, value in d.items():
        if isinstance(value, dict):
            target[key] = _filter_null_keys(value)
        elif value:
            target[key] = value

    return target



import re
from setuptools import find_packages, setup


def read_requirements(req_file):
    return [l for l in re.sub(r"\s*#.*\n", "\n", req_file.read()).splitlines() if l]


with open("requirements/base.txt") as f:
    REQUIREMENTS = read_requirements(f)

with open("requirements/test.txt") as f:
    TEST_REQUIREMENTS = read_requirements(f)

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="connexion-compose",
    version="0.3.0",
    description="Create Connexion schema composed from several files in a nested directory structure.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/jsmesami/connexion-compose",
    author="Ondřej Nejedlý",
    author_email="jsmesami@gmail.com",
    license="MIT",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    include_package_data=True,
    keywords="connexion swagger schema compose",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.8",
    ],
)



from decimal import Decimal
import re
from urllib.parse import urlparse

from jsonschema import draft4_format_checker

ISO_4217_RE = re.compile("^[A-Z]{3}$")

DECIMAL_RE = re.compile("^[0-9]{0,20}.?[0-9]{0,6}$")

RFC_3339_RE = re.compile(
    r"^(\d+)-(0[1-9]|1[012])-(0[1-9]|[12]\d|3[01])[Tt]"
    r"([01]\d|2[0-3]):([0-5]\d):([0-5]\d|60)(\.\d+)?(([Zz])|([+|\-]([01]\d|2[0-3]):[0-5]\d))$"
)

UUID_RE = re.compile("^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$", re.IGNORECASE)


@draft4_format_checker.checks("iso_4217")
def is_iso_4217(val):
    """Currency code validator."""
    if not isinstance(val, str):
        return False
    return bool(ISO_4217_RE.match(val))


@draft4_format_checker.checks("decimal")
def is_decimal(val):
    if isinstance(val, Decimal):
        return True
    if not isinstance(val, str):
        return False
    return bool(DECIMAL_RE.match(val))


@draft4_format_checker.checks("rfc_3339")
def is_rfc_3339(val):
    """Timestamp validator."""
    if not isinstance(val, str):
        return False
    return bool(RFC_3339_RE.match(val))


@draft4_format_checker.checks("url")
def is_url(val):
    try:
        if not isinstance(val, str):
            return False
        parsed = urlparse(val)
        if parsed.scheme not in {"http", "https"}:
            return False
        if not parsed.netloc:
            return False
        if parsed.port:
            pass
        return True
    except ValueError:
        return False


@draft4_format_checker.checks("path")
def is_path(val):
    try:
        if not isinstance(val, str):
            return False
        parsed = urlparse(val)
        if parsed.scheme or parsed.netloc:
            return False
        return bool(parsed.path)
    except ValueError:
        return False


@draft4_format_checker.checks("uuid")
def is_uuid(val):
    if not isinstance(val, str):
        return False
    return bool(UUID_RE.match(val))
